{% extends 'allianceauth/base.html' %}
{% load i18n %}
{% load static %}

{% block page_title %}{% translate app_title %}{% endblock %}
{% block content %}
    <div class="col-lg-12">
        <h2 class="page-header text-center">{% translate app_title %}</h2>
        <div class="col-lg-10 col-lg-offset-1 container">
            {% block details %}{% endblock %}
        </div>
    </div>
{% endblock %}

{% block extra_css %}
    <link
        href="{% static 'standingssync/base.css' %}"
        type="text/css"
        rel="stylesheet"
    >
{% endblock %}
