# Generated by Django 4.0.6 on 2022-07-29 17:08

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("standingssync", "0003_war_target_labels"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="evewar",
            name="aggressor",
        ),
        migrations.RemoveField(
            model_name="evewar",
            name="allies",
        ),
        migrations.RemoveField(
            model_name="evewar",
            name="defender",
        ),
        migrations.DeleteModel(
            name="EveContact",
        ),
        migrations.DeleteModel(
            name="EveEntity",
        ),
        migrations.DeleteModel(
            name="EveWar",
        ),
    ]
